#ifndef PATHNAMES_H
#define PATHNAMES_H


/* Lua imapfilter set functions file. */
#define PATHNAME_COMMON			CONFIG_SHAREDIR "/common.lua"

/* Lua imapfilter set functions file. */
#define PATHNAME_SET			CONFIG_SHAREDIR "/set.lua"

/* Lua imapfilter account functions file. */
#define PATHNAME_ACCOUNT		CONFIG_SHAREDIR "/account.lua"

/* Lua imapfilter mailbox functions file. */
#define PATHNAME_MAILBOX		CONFIG_SHAREDIR "/mailbox.lua"

/* Lua imapfilter message functions file. */
#define PATHNAME_MESSAGE		CONFIG_SHAREDIR "/message.lua"

/* Lua imapfilter message functions file. */
#define PATHNAME_OPTIONS		CONFIG_SHAREDIR "/options.lua"

/* Lua imapfilter regex functions file. */
#define PATHNAME_REGEX			CONFIG_SHAREDIR "/regex.lua"

/* Lua imapfilter auxiliary functions file. */
#define PATHNAME_AUXILIARY		CONFIG_SHAREDIR "/auxiliary.lua"


#endif				/* PATHNAMES_H */
