#ifndef VERSION_H
#define VERSION_H


/* Program's version number. */
#define VERSION		"2.8.5"

/* Program's copyright. */
#define COPYRIGHT	"Copyright (c) 2001-2026 Eleftherios Chatzimparmpas"


#endif				/* VERSION_H */
